from __future__ import annotations
from functools import wraps
from textwrap import dedent
'''
Args functions:
    For reduction functions whose return values are
    **indices** or **coordinates** over multiple axes
    E.g. argsmax
'''

from typing import Protocol

import numpy as np
import numpy.typing as npt

from .. import numpy as xtnp
from .. import tensor as xtt

from ..tensor import XTensor, TensorLike, DimLike, DimsLike


ARGS_DIM = 'ARGS_DIM'

class _args_func(Protocol):
    def __call__(self, a: np.ndarray, axes: xtnp.AxesLike) -> npt.NDArray[np.int_]: ...


class ArgsFunction(Protocol):
    """
    Argument function protocol
    """
    def __call__(self, x: TensorLike, /, dim: DimsLike) -> XTensor:
        """
        :param x: target tensor
        :param dim: target dimension(s)

        :return: an :py:class:`XTensor` of rank :code:`x.rank - len(dim) + 1`

        """
        ...


def _reduction_factory(_func: _args_func) -> ArgsFunction:
    @xtt.generalize_at_0
    def _reduce(X: xtt.XTensor, /, dim: xtt.DimsLike) -> xtt.XTensor:
        '''
        Return
        '''

        axes = X.get_axes(dim)
        r_dims, s_dims = xtt.strip(X.dims, axes, only_remaining=False)
        r_coords, s_coords = xtt.strip(X.coords, axes, only_remaining=False)

        new_dims = r_dims + [ARGS_DIM]
        new_coords = r_coords + [[dim for dim in s_dims]]

        args = _func(X.data, axes=axes)
        return xtt.XTensor(args, dims=new_dims, coords=new_coords)
    return _reduce


def _inject_docs(r: ArgsFunction, doc: str):
    r.__doc__ = dedent(doc)
    return r

def _inject_sig(r: ArgsFunction):
    def _dummy(x: TensorLike, /, dim: DimLike|DimsLike|None) -> XTensor:
        ...
    _dummy.__doc__ = r.__doc__
    return wraps(_dummy)(r)

def postproc(doc: str):
    def _postproc(r: ArgsFunction):
        return _inject_sig(_inject_docs(r, doc))

    return _postproc


argsmin = postproc("""
                    Return the indices where minima occur.
                   """)(_reduction_factory(xtnp.argsmin))


argsmax = postproc("""
                    Return the indices where maxima occur.
                    """)(_reduction_factory(xtnp.argsmax))

nanargsmin = postproc("""
                      Similar to :py:func:`argsmin`, but with :code:`nan` ignored
                      """)(_reduction_factory(xtnp.nanargsmin))

nanargsmax = postproc("""
                      Similar to :py:func:`argsmax`, but with :code:`nan` ignored
                      """)(_reduction_factory(xtnp.nanargsmax))






from __future__ import annotations
from functools import wraps
from textwrap import dedent
from typing import Protocol


import numpy as np
import numpy.typing as npt
from .. import tensor as xtt
from ..tensor import XTensor

'''
Arg functions:
    For single-dimension reduction functions whose return values are array
    **indices** or **coordinates**
    E.g. argmax
'''

class _np_arg_func(Protocol):
    def __call__(self, a: np.ndarray, axis: int) -> npt.NDArray[np.int_]: ...

class ArgFunction(Protocol):
    def __call__(self, x: xtt.TensorLike, /, dim: xtt.DimLike) -> xtt.XTensor: ...

class CoordFunction(Protocol):
    def __call__(self, x: xtt.TensorLike, /, dim: xtt.DimLike, *, 
            use_index_if_no_coord: bool=False) -> xtt.XTensor: ...


def _reduction_factory(_np_func: _np_arg_func) -> ArgFunction:
    @xtt.generalize_at_0
    def _reduce(X: xtt.XTensor, /, dim: xtt.DimLike) -> xtt.XTensor:
        axis = X.get_axis(dim)
        return xtt.XTensor(
                _np_func(X.data, axis=axis),
                dims=xtt.strip(X.dims, [axis]),
                coords=xtt.strip(X.coords, [axis]))
    return _reduce


def _inject_docs(r: ArgFunction, doc: str):
    r.__doc__ = dedent(doc)
    return r

def _inject_sig(r: ArgFunction):
    def _dummy(x: xtt.TensorLike, /, dim: xtt.DimLike|xtt.DimsLike|None) -> XTensor:
        ...
    _dummy.__doc__ = r.__doc__
    return wraps(_dummy)(r)

def postproc(doc: str):
    def _postproc(r: ArgFunction):
        return _inject_sig(_inject_docs(r, doc))

    return _postproc


_argmax = postproc("""
                    :return: 
                   """)(_reduction_factory(np.argmax))

_argmin = _reduction_factory(np.argmin)

_nanargmax = _reduction_factory(np.nanargmax)
_nanargmin = _reduction_factory(np.nanargmin)


def _coord_reduc_factory(_func: ArgFunction) -> CoordFunction:
    @xtt.generalize_at_0
    def _reduce(X: xtt.XTensor, /, dim: xtt.DimLike, *, use_index_if_no_coord: bool=False) -> xtt.XTensor:
        '''
            Return the coordinate on [dim] that maximizes/minimizes x If dimension [dim] does
            not have coordinates, this is equivalent to (nan)argmax/argmin
        '''
        args = _func(X, dim)
        axis = X.get_axis(dim)
        coord = X.coords[axis]

        if coord is None: 
            if use_index_if_no_coord:
                coord_r = np.arange(X.shape[axis])[args]
            else:
                raise ValueError(f'Tensor has no coordinates on axis {axis}')
        else:
            coord_r = coord[args]

        return xtt.XTensor(coord_r, dims=xtt.strip(X.dims, [axis]), coords=xtt.strip(X.coords, [axis]))
    return _reduce


_coordmax = _coord_reduc_factory(_argmax)
_coordmin = _coord_reduc_factory(_argmin)

_nancoordmax = _coord_reduc_factory(_nanargmax)
_nancoordmin = _coord_reduc_factory(_nanargmin)






from ._base import XTensor

from ._decors import promote_binary_operator, promote_ternary_operator

from ._slice import TensorSlice, MetaTensorSlice

from .broadcast import *

from .basic_utils import *

from .typing import *



from __future__ import annotations
from functools import wraps
from textwrap import dedent
'''
Args functions:
    For reduction functions whose return values are
    **indices** or **coordinates** over multiple axes
    E.g. argsmax
'''

from typing import Protocol

import numpy as np
import numpy.typing as npt

from .. import numpy as xtnp
from .. import tensor as xtt

from ..tensor import XTensor, TensorLike, DimLike, DimsLike


from ._args import ArgsFunction, argsmin, argsmax, nanargsmin, nanargsmax


class CoordsFunction(Protocol):
    """
    Coordinate function protocol
    """
    def __call__(self, x: xtt.TensorLike, /, dim: xtt.DimsLike, *, 
            use_index_if_no_coord: bool=False) -> xtt.XTensor:
        """
        :param x: target tensor
        :param dim: target dimension(s)
        :param use_index_if_no_coord: whether to use indices instead of
                        coordinates are not specified

        :return: an :py:class:`XTensor` of rank :code:`x.rank - len(dim) + 1`
        """
        ...


def _coord_reduc_factory(_func: ArgsFunction) -> CoordsFunction:
    @xtt.generalize_at_0
    def _reduce(X: xtt.XTensor, /, dim: xtt.DimsLike, *, use_index_if_no_coord: bool=False) -> xtt.XTensor:
        '''
        Return the coordinate on [dim] that maximizes/minimizes x If dimension [dim] does
        not have coordinates, this is equivalent to (nan)argmax/argmin
        '''
        args = _func(X, dim)
        axes = X.get_axes(dim)

        coords = [X.coords[axis] for axis in axes]
        coords_r = []

        for i, (axis, coord) in enumerate(zip(axes, coords)):

            if coord is None: 
                if use_index_if_no_coord:
                    coord_ = np.arange(X.shape[axis])
                else:
                    raise ValueError(f'Tensor has no coordinates on axis {axis}')
            else:
                coord_ = coord


            coords_r.append(coord_[args.data[...,i]])

        return xtt.XTensor(
                np.stack(coords_r, axis=-1),
                dims=args.dims, 
                coords=args.coords)

    return _reduce

def _inject_docs(r: CoordsFunction, doc: str):
    r.__doc__ = dedent(doc)
    return r

def _inject_sig(r: CoordsFunction):
    def _dummy(x: TensorLike, /, dim: DimLike|DimsLike|None, *,
                use_index_if_no_coord: bool=False
        ) -> XTensor:
        ...
    _dummy.__doc__ = r.__doc__
    return wraps(_dummy)(r)

def postproc(doc: str):
    def _postproc(r: CoordsFunction):
        return _inject_sig(_inject_docs(r, doc))

    return _postproc


coordsmin = postproc("""
                     Return the coordinates where minima occur
                     """)(_coord_reduc_factory(argsmin))

coordsmax = postproc("""
                     Return the coordinates where maxima occur
                     """)(_coord_reduc_factory(argsmax))

nancoordsmin = postproc("""
                        Similar to :py:func:`coordsmin`, but with :code:`nan` ignored
                        """)(_coord_reduc_factory(nanargsmin))

nancoordsmax = postproc("""
                        Similar to :py:func:`coordsmax`, but with :code:`nan` ignored
                        """)(_coord_reduc_factory(nanargsmax))





from __future__ import annotations
from typing import TYPE_CHECKING, Any, List, Protocol, Sequence, Tuple
import numpy as np


if TYPE_CHECKING:
    from ...tensor._base import XTensor
    from ..typing import AxesPermutation, Coords, Dims


class Dimcaster(Protocol):
    """
    The dimension caster protocol.
    """
    def __call__(self, X: XTensor, Y: XTensor) -> Tuple[AxesPermutation, AxesPermutation]:
        r"""
        :param x,y: A pair of :py:class:`xtensors.XTensor` objects 

        :return: A pair of :py:class:`xtensors.AxesPermutation` (i.e. lists of ints) 
                 specifying how the axes of :code:`X` and :code:`Y` should be permuted.
                 
        """
        ...


class DimMerger(Protocol):
    """
    Dimension-name-merging protocol
    """
    def __call__(self, X: XTensor, Y: XTensor) -> Dims:
        """
        Merge the dimension names of :code:`X` and :code:`Y`

        :rtype: :py:class:`xtensors.Dims`
         
        """
        ...


class CoordMerger(Protocol):
    """
    Coordinates-mergin protocol
    """
    def __call__(self, X: XTensor, Y: XTensor) -> Coords:
        """
        Merge the coordinates of of :code:`X` and :code:`Y`

        :rtype: :py:class:`xtensors.Coords`

        """
        ...


class Broadcaster(Protocol):
    """
    Broadcaster protocol
    """
    def __call__(
        self,
        X: XTensor, Y: XTensor
    ) -> Tuple[np.ndarray, np.ndarray, Dims, Coords]:
        """
        Broadcast two tensors together

        :return: The resulting :code:`np.ndarray` s that are mutually
                broadcastable, and the new dimension names as well as the coordinates
                after broadcasting.

        """
        ...









from typing import TYPE_CHECKING as __TYPE_CHECKING

from ._typing import TensorLike, DimsLike, HasDimName, DimLike, Array, AxesPermutation, Dims, Coords

if __TYPE_CHECKING:
    from ._typing import AxesPermutation, BinaryOperator, TernaryOperator
    from ._typing import Function_2Args, Function_1Arg, Function_3Args



from __future__ import annotations
from functools import wraps
from textwrap import dedent
from typing import Protocol, Tuple
import numpy as np

from .. import tensor as xtt
from ..tensor import XTensor, TensorLike, DimLike, DimsLike


'''
For reduction functions that can act over multiple axes/dims.
'''

class _np_reduction_func(Protocol):
    def __call__(self, a: np.ndarray, axis: int|Tuple[int,...]) -> np.ndarray: ...


class ReductionFunc(Protocol):
    def __call__(self, x: xtt.TensorLike,/, dim: xtt.DimLike|xtt.DimsLike|None) -> xtt.XTensor: ...


def _reduction_factory(_np_func: _np_reduction_func) -> ReductionFunc:
    @xtt.generalize_at_0
    def _reduce(X: xtt.XTensor, /, dim: xtt.DimLike|xtt.DimsLike|None=None) -> xtt.XTensor:
        axes = X.get_axes(dim)

        _y = _np_func(X.data, axis=tuple(axes))

        return xtt.XTensor(_y, dims=xtt.strip(X.dims, axes), coords=xtt.strip(X.coords, axes))
    return _reduce


def _inject_docs(r: ReductionFunc, doc: str):
    r.__doc__ = dedent(doc)
    return r

def _inject_sig(r: ReductionFunc):
    def _dummy(x: TensorLike, /, dim: DimLike|DimsLike|None) -> XTensor:
        ...

    _dummy.__doc__ = r.__doc__
    return wraps(_dummy)(r)

def postproc(doc: str):
    def _postproc(r: ReductionFunc):
        return _inject_sig(_inject_docs(r, doc))

    return _postproc



_sum = postproc(r"""
                :return: :math:`\sum_{\mathrm{dim}} x`

                """)(_reduction_factory(np.sum))

_mean = postproc(r"""
                :return: :math:`\braket{x}_\mathrm{dim}`
                 """)(_reduction_factory(np.mean))


_std = postproc(r"""
                :return: :math:`\sqrt{\braket{x^2}_\mathrm{dim} - \braket{x}_\mathrm{dim}^2}`
                """)(_reduction_factory(np.std))

_nanmean = postproc(r"""
                    Same as :py:meth:`xtensors.mean`, but :code:`nan` is ignored
                    """)(_reduction_factory(np.nanmean))

_nanstd = postproc(r"""
                   Same as :py:meth:`xtensors.std`, but :code:`nan` is ignored
                    """)(_reduction_factory(np.nanstd))

_nansum = postproc(r"""
                   Same as :py:meth:`xtensors.sum`, but :code:`nan` is ignored
                    """)(_reduction_factory(np.nansum))


_max = postproc(r"""
                :return: :math:`\max_\mathrm{dim} x`
                """)(_reduction_factory(np.max))

_min = postproc(r"""
                :return: :math:`\min_\mathrm{dim} x`
                """)(_reduction_factory(np.min))

_nanmax = postproc(r"""
                Same as :py:meth:`xtensors.max`, but :code:`nan` is ignored
                """)(_reduction_factory(np.nanmax))

_nanmin = postproc(r"""
                Same as :py:meth:`xtensors.min`, but :code:`nan` is ignored
                """)(_reduction_factory(np.nanmin))

_all = postproc(r"""
                Used on bool-valued :code:`TensorLike` objects,
                similar to :code:`np.all`
                """)(_reduction_factory(np.all))


_any = postproc(r"""Used on bool-valued :code:`TensorLike` objects,
                similar to :code:`np.any`
                """)(_reduction_factory(np.any))








from __future__ import annotations
from functools import wraps
from typing import Callable, Protocol
from textwrap import dedent


import numpy as np

from .. import tensor as xtt

from ..tensor import XTensor, TensorLike



class _ufunc(Protocol):
    def __call__(self, __x1: np.ndarray, /, *args) -> np.ndarray: ...


class UFunc(Protocol):
    def __call__(self, x: xtt.TensorLike, /) -> xtt.XTensor: ...


def _ufunc_factory(_np_func: _ufunc) -> UFunc:
    @xtt.generalize_at_0
    def _f(X: xtt.XTensor, /) -> xtt.XTensor:
        return xtt.XTensor(_np_func(X.data), X.dims, X.coords)
    return _f


def _inject_docs(ufunc: UFunc, func_name: str) -> UFunc:
    _docs = dedent(f"""
        :param x: Input tensor

        :return: :math:`{func_name}(x)`

    """)
    ufunc.__doc__ = _docs
    return ufunc


def _inject_sig(ufunc: UFunc) -> UFunc:
    def _dummy(x: TensorLike, /) -> XTensor:
        ...
    _dummy.__doc__ = ufunc.__doc__

    return wraps(_dummy)(ufunc)


def _sigmoid(__x1: np.ndarray):
    return .5 * (1. + np.tanh(__x1/2))


def postproc(*args):
    def _postproc(f: UFunc):
        return _inject_sig(_inject_docs(f, *args))

    return _postproc


# sigmoid = _inject_signature(_inject_docs(_ufunc_factory(_sigmoid), r'\mathrm{sigmoid}'))

sigmoid = postproc(r'\mathrm{sigmoid}')(_ufunc_factory(_sigmoid))

exp = postproc(r'\exp')(_ufunc_factory(np.exp))

log2 = postproc(r'\log_2')(_ufunc_factory(np.log2))

log = postproc(r'\ln')(_ufunc_factory(np.log))

log10 = postproc(r'\log_{10}')(_ufunc_factory(np.log10))

cos = postproc(r'\cos')(_ufunc_factory(np.cos))

sin = postproc(r'\sin')(_ufunc_factory(np.sin))

tan = postproc(r'\tan')(_ufunc_factory(np.tan))

cosh = postproc(r'\cosh')(_ufunc_factory(np.cosh))

sinh = postproc(r'\sinh')(_ufunc_factory(np.sinh))

tanh = postproc(r'\tanh')(_ufunc_factory(np.tanh))








from __future__ import annotations
from functools import wraps
from textwrap import dedent

from .. import tensor as xtt

from ..tensor import XTensor, TensorLike

from typing import Optional, Protocol
import numpy as np
import numpy.typing as npt

import warnings


class BinaryOperation(Protocol):
    def __call__(self, x: xtt.TensorLike, y: xtt.TensorLike, /) -> xtt.XTensor: ...


def _apply_operation(X: xtt.XTensor, Y: xtt.XTensor, binop: str, rbinop: Optional[str]=None) -> xtt.XTensor:
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', 'invalid value encountered in true_divide')
        try:
            Z = getattr(X, binop)(Y)
            if Z is NotImplemented: raise NotImplementedError

        except (AttributeError, NotImplementedError) as e:
            if rbinop is not None:
                Z = getattr(Y, rbinop)(X)
                if Z is NotImplemented: raise NotImplementedError
            else:
                raise AttributeError from e
    return Z


def _binop_factory(_bin_op: str, _rbin_op: str) -> BinaryOperation:
    @xtt.generalize_at_1
    @xtt.generalize_at_0
    def _op(X: xtt.XTensor, Y: xtt.XTensor, /) -> xtt.XTensor:
        return _apply_operation(X, Y, _bin_op, _rbin_op)

    return _op


def _inject_docs(b: BinaryOperation, operation: str) -> BinaryOperation:
    
    b.__doc__ = dedent(
        fr"""
            :param x,y: :py:class:`xtensors.TensorLike` objects

            :return: :math:`{operation}`
        """
    )

    return b


def _inject_sig(b: BinaryOperation) -> BinaryOperation:
    def _dummy(x: TensorLike, y: TensorLike, /) -> XTensor:
        ...

    _dummy.__doc__ = b.__doc__
    return wraps(_dummy)(b)


def postproc(operation: str):
    def _postproc(b: BinaryOperation):
        return _inject_sig(_inject_docs(b, operation))

    return _postproc

_add = postproc('x + y')(_binop_factory('__add__', '__radd__'))

_divide = postproc('x / y')(_binop_factory('__truediv__', '__rtruediv__'))

_multiply = postproc('xy')(_binop_factory('__mul__', '__rmul__'))

_greater = postproc('x > y')(_binop_factory('__gt__', '__lt__'))

_greater_equal = postproc(r'x \ge y')(_binop_factory('__ge__', '__le__'))

_less = postproc(r'x < y')(_binop_factory('__lt__', '__gt__'))

_less_equal = postproc(r'x \le y')(_binop_factory('__le__', '__ge__'))

_equal = postproc(r'x = y')(_binop_factory('__eq__', '__eq__'))


def _np_or(X: npt.NDArray, Y: npt.NDArray) -> npt.NDArray:
    return np.logical_or(X, Y)

def _np_and(X: npt.NDArray, Y: npt.NDArray) -> npt.NDArray:
    return np.logical_and(X, Y)




_or = postproc(r'x\;\mathrm{or}\;y')(
        xtt.generalize_at_0(
        xtt.generalize_at_1(
        xtt.promote_binary_operator(xtt.vanilla_broadcaster)(
            _np_or
))))


_and = postproc(r'x\;\mathrm{and}\;y')(
        xtt.generalize_at_0(
        xtt.generalize_at_1(
        xtt.promote_binary_operator(xtt.vanilla_broadcaster)(
            _np_and
))))





from __future__ import annotations
from abc import abstractmethod

'''
Template Broadcaster:

    First specify a template, then map axes of tensors onto the template

'''

import numpy as np
from ..basic_utils import permute, permutation_well_defined, mergecoords, mergedims

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import List, Literal, Optional, Sequence, Union
    from .._base import XTensor
    from ..typing import Dims, Coords

    Channel = Union[str, None]


class AxisSelector:
    """
    Abstract base class for objects that implement :code:`select_axis`
    
    """
    def __init__(self) -> None:
        self._label: str
        self.channel: str|None

    @abstractmethod
    def select_axis(self, X: XTensor, channel: str|None=None) -> int|None:
        """
        Select an axis from the tensor X
        
        :param X: tensor
        :param channel: (optional) a string specifying the channel. If
                :code:`self.channel` is set, 
                axis selection will only be performed
                when the :code:`channel` parameter
                matches :code:`self.channel`, or else :code:`None` is returned.
        """

    @property
    def label(self) -> str:
        return self._label


class DimNameSelector(AxisSelector):
    """
    Select an axis based on its name

    """
    def __init__(
        self, dimname: str, 
        required: bool=False, channel: Optional[str]=None) -> None:
        """
        :param dimname: target axis name
        :param required: if :code:`True`, a :code:`KeyError` will be raised if 
                no axis named :code:`dimname` is found when selecting an axis from a tensor.

        :param channel: Selecting channel.

        """
        self.dimname = dimname
        self.required = required
        self.channel = channel
        self._label = dimname

    def select_axis(self, X: XTensor, channel: str|None=None) -> int|None:
        if self.channel is not None and channel != self.channel:
            return None
        try:
            return X.get_axis(self.dimname)
        except KeyError as e:
            if self.required: raise e
            return None



class IndexSelector(AxisSelector):
    """
    Select an axis based on the index

    """
    def __init__(self, axis: int, channel: Optional[str]=None) -> None:
        """
        :param axis: The axis index, negative indices are supported
        :param channel: Selecting channel.

        """
        self.axis = axis
        self.channel = channel
        self._label = str(axis)


    def select_axis(self, X: XTensor, channel: str|None=None) -> int | None:
        if self.channel is not None and channel != self.channel:
            return None
        
        if self.axis >= 0:
            return self.axis
        else:
            return self.axis + X.rank


class Template:
    """
    A tensor broadcasting template that can be reused to broadcast multiple
    tensors.

    A template is basically a list of :py:class:`AxisSelector` instances, each
    specifying how an axis (integer) should be selected.

    """
    def __init__(self, *selectors: AxisSelector) -> None:
        """
        :param selectors: A list of :py:class:`AxisSelector`
        """

        try:
            for sel in selectors: assert isinstance(sel, AxisSelector)
        except AssertionError as e:
            raise TypeError('selectors must be instances of AxisSelector') from e

        self.selectors = selectors
        self._dims: Dims
        self._coords: Coords
        self.clear()
        self.check_well_defined()


    @classmethod
    def from_dims_channels(
            cls,
            dims: Sequence[Union[str, int]],
            channels: Sequence[str|None]|None=None) -> Template:
        selectors = []

        if channels is None: channels = [None for _ in range(len(dims))]

        assert len(dims) == len(channels)

        for d, c in zip(dims, channels):
            if isinstance(d, str):
                selectors.append(DimNameSelector(d, channel=c))
            else:
                selectors.append(IndexSelector(d, channel=c))
        
        return cls(*selectors)

    def check_well_defined(self):
        """
        Check if the :code:`AxisSelector` s are configured in a self-consistent
        manner.

        In particular, this method checks if there are :code:`DimNameSelector` s with 
        duplicate names or :code:`AxisSelector` s with duplicate indices.

        :raises: :code:`ValueError` if the check fails

        """
        flag = False
        names = [sel.dimname for sel in self.selectors if isinstance(sel, DimNameSelector)]

        if len(names) != len(set(names)):
            flag = True

        axes = [sel.axis for sel in self.selectors if isinstance(sel, IndexSelector)]
        if len(axes) != len(set(axes)):
            flag = True

        if flag:
            raise ValueError('Axis selectors not consistent')

    def cast_and_update(self, X: XTensor, channel: str|None=None) -> XTensor:
        """
        Cast the input tensor according to the template and update self's
        internal state (i.e. dimension names and coordinates).

        """
        axesp = [sel.select_axis(X, channel) for sel in self.selectors]
        assert permutation_well_defined(axesp, X.rank), 'Casting impossible'

        X1 = permute(X, axesp)

        self._dims = mergedims(self._dims, X1)
        self._coords = mergecoords(self._coords, X1)

        return X1

    def clear(self):
        """
        Reset the internal states.

        """
        self._dims = [None for _ in range(len(self.selectors))]
        self._coords= [None for _ in range(len(self.selectors))]

    @property
    def dims(self) -> Dims:
        """
        The dimension names gathered from casting tensors.
        """
        return self._dims

    @property
    def coords(self) -> Coords:
        """
        The coordinates names gathered from casting tensors.
        """

        return self._coords

    def __repr__(self) -> str:
        s = 'Dimension Template'
        s1 = 'dim.    '
        s2 = 'channel '

        for selector in self.selectors:
            label = selector.label
            channel = str(selector.channel)

            width = max(len(label), len(channel)) + 1

            s1 += f'{label:{width}}'
            s2 += f'{channel:{width}}'


        return s + '\n' + s1 + '\n' + s2




from __future__ import annotations
from functools import wraps


from typing import Callable, List, Literal, Sequence, Tuple, TypeVar, overload


T = TypeVar('T')

@overload
def strip(l: Sequence[T], indices: Sequence[int]) -> List[T]: ...
@overload
def strip(l: Sequence[T], indices: Sequence[int], only_remaining: Literal[True]) -> List[T]: ...
@overload
def strip(l: Sequence[T], indices: Sequence[int], only_remaining: Literal[False]) -> Tuple[List[T], List[T]]: ...

def strip(l: Sequence[T], indices: Sequence[int], only_remaining: bool=True) -> List[T]|Tuple[List[T], List[T]]:
    remaining, stripped = [], []

    for i, x in enumerate(l):
        if i in indices: stripped.append(x)
        else: remaining.append(x)

    if only_remaining: return remaining
    return remaining, stripped



def copy_sig(f: Callable):
    def wrapper(g: T) -> T:
        _doc = g.__doc__
        g = wraps(f)(g)
        g.__doc__ = _doc
        return g

    return wrapper


def copy_doc(f: Callable):

    def wrapper(g: T) -> T:
        g.__doc__ = f.__doc__
        return g
    return wrapper




from .base import *

from .tensor import *

from .version import __version__



# from .unify import (
#         rename, 
#         strip_dims, get_axis_from_dim, get_axes, 
#         new_axes, new_axis
#         )
#
#
#
